"""
    Runs a transliteration network, choosing specific parameters from the command line.
"""
import sys
from sys import argv, stderr, exit
import argparse
from translit_lib.vocabulary import Vocabulary
from translit_lib.corpus_utils import BatchCorpus, create_and_train, computeRepresentation
from translit_lib.encoder_decoder import Encoder, Decoder, BidiEncoder, Transliterator
from translit_lib.beam import Beam

# Current corpus folder for running the tests.

corpus_folder = "data"

def processArgs():
    usage= '''Transliterate using some network we have built.
    options marked as 'S' in the help only apply for single sentences given by --mdc.
    '''
    parser = argparse.ArgumentParser(description=usage)
    parser.add_argument("--net", dest="net", help="path of the network to use", required=True)
    parser.add_argument("--arch",  dest="arch", help="network architecture : plain, bidi, seg or bidiseg", default="bidiseg")
    parser.add_argument("--hidden",  dest="hidden", help="size of hidden layer", default=500, type=int)
    parser.add_argument("--beam",  dest="beam", help="beam size", default=1, type=int)
    parser.add_argument("--corpus",  dest="corpus", help="corpus to transliterate: val, test, htest, ctest", default="val") 
    parser.add_argument("--mdc",  dest="mdc", help="sentence to transliterate")
    parser.add_argument("--compact",
         help="if not set, all characters will be separated by spaces (useful for batch and test)",
                    action="store_true")
    parser.add_argument("--attention",  dest="attention_pict", help="filename for attention map (S)")
    parser.add_argument("--verbose", help="verbose output : (NO-OP currently)", 
        action="store_true")
    parser.add_argument("--display_attention", help="display attention", 
        action="store_true")
    
    if len(argv)==1:
        parser.print_help(stderr)
        exit(1)
    args = parser.parse_args()
    return vars(args)

def run_network(net,corpus="test",mdc=None,compact=False, hidden=300, arch="plain", 
    beam=1, verbose=False, attention_pict=None, display_attention=False):
    vocabulary = Vocabulary()
    vocabulary.loadDefault()

    # Boolean indicator needed to state we want to keep attention information.
    attention_output=(attention_pict != None) or display_attention

    if arch == "bidi" or arch == "bidiseg" :
        encoder = BidiEncoder(input_size=vocabulary.size(), hidden=hidden)
        encoder_model = encoder.load_model_from_file(net)
        decoder = Decoder(hidden=hidden)
        decoder_model = decoder.load_model_from_file(net, attention_output=attention_output)
        pass
    elif arch == "seg" or arch == "plain":
        encoder = Encoder(input_size=vocabulary.size(), hidden=hidden)
        encoder_model = encoder.load_model_from_file(net)
        decoder = Decoder(hidden=hidden)
        decoder_model = decoder.load_model_from_file(net)
        pass
    else:
        print("unknown arch "+ arch)
        exit(1)
   
    transliterator = Transliterator(vocabulary, encoder_model, decoder_model, attention_output=attention_output)
    
    if mdc :
        trl = transliterator.transliterate(mdc, beam_width=beam)
        print(trl)
        if attention_pict:
            transliterator.plot_attention(picture=attention_pict)
        if display_attention:
            transliterator.plot_attention()
    else :
        corpus_file = "../{}/src-{}.txt".format(corpus_folder,corpus)
        with open(corpus_file) as file:
            for l in file:
                trl = transliterator.transliterate(l, beam_width=beam)
                if not compact:
                    trl = " ".join(trl.replace(" ", "_"))
                print(trl)


if __name__ == "__main__":
    params=processArgs()
    run_network(**params)
