"""
Generic tools used by all networks.
"""

import random
import os
import sys
import operator
import numpy as np
import re

import matplotlib.pyplot as plt

from keras.callbacks import ModelCheckpoint
from keras.preprocessing import sequence
from keras.models import load_model
from keras.preprocessing import sequence
from keras.utils import plot_model, to_categorical
import keras.backend as K

from translit_lib.vocabulary import Vocabulary


def clean_mdc_input(string):
    """
    Cleanup a MDC String, removing non-codes.
    (not 100% safe, BTW, don't use it on real MDC)
    """
    return re.sub("[^a-zA-Z0-9]+", " ", string) # clean up input
    
# Basic methods
def encode_mdc_string(string, vocabulary):
    """
    Converts space-separated MDC codes to a one-hot numpy tensor.
    """
    list = vocabulary.encode_string(string)
    return computeRepresentation([list], num_classes= vocabulary.size())  

def computeRepresentation(data, num_classes=256):
    '''
    Compute a binary representation for text
    - text : list of list of codes
    - returns a numpy one hot array of size BATCH x LENGTH x num_classes
    '''
    tab=[to_categorical(line, num_classes=num_classes) for line in  data]
    res = np.array(sequence.pad_sequences(tab, padding='post'), dtype=float)
    return res
    
def decodeRepresentation(array):
    """
    Takes a numpy softmax array of size BATCH x LENGTH x num_classes
    and returns the most likely (greedy chain).
    """
    res= []
    for i in range(array.shape[0]):
        line = ""
        for j in range(array.shape[1]):
            t = np.argmax(array[i][j]) 
            line = line + chr(t)
        res.append(line)
    return res

# TODO : reorganize this...
def _load_character_corpus(file_name, max_code=255, space_separated=True):
    """
    Loads a corpus made of characters, and encode it using character codes.
    - codes greater than max_codes are mapped to 3 (unknown)
    - limits (1 and 2) are included
    - if space_separated is true, actual chars are separated from each other by a meaningless space
    """
    def cutCode(code, max_code):
        return code if code <= max_code else 3
    res = []
    with open(file_name, encoding="utf-8") as f:
        for line in f:
            line = line.strip("\n\r\t ")
            if space_separated:
                line = line.split(" ")
            codes= [1] + [cutCode(ord(c), max_code) for c in line] + [2]
            res.append(codes)
    return res

# From numeric lists of lists to numpy one-hot-arrays...

def _corpus_list_to_np(data, num_classes):
    """
    An utility low level function to transform lists of lists of data into numpy array.
    """
    # batch_size = len(data)
    tab=[to_categorical(line, num_classes=num_classes) for line in data]
    res = np.array(sequence.pad_sequences(tab, padding='post'), dtype=float)
    return res
    
def _encode_corpus_for_seq2seq(
    encoded_inputs, encoded_outputs,
    num_input_classes, num_output_classes):
    """
    propertly prepare encoded batch entry for seq2seq processing.
    Useful in generators.
    The entries are simple lists of lists of codes. The function will 
    add to them the specific codes for sentence limits (1 and 2) where needed.
    - encoded_inputs: lists of lists of entry codes (integers)
    - encoded_outputs: lists of lists of entry codes (integers)
    - num_input_classes: number of classes of input entries
    - num_input_classes: number of classes of outputs
    - returns a complex 2-uple of values ((a,b),c)
        a,b are numpy arrays for encoder and decoder inputs
        c is a numpy array for the expected output
    """
    numpy_encoder_input = _corpus_list_to_np(encoded_inputs, num_classes=num_input_classes)
    # base for the decoder...
    numpy_decoder_in_out = _corpus_list_to_np(encoded_outputs, num_classes=num_output_classes)
    numpy_decoder_input = numpy_decoder_in_out[:,:-1,:]
    numpy_decoder_output= numpy_decoder_in_out[:,1:,:]
    return ([numpy_encoder_input, numpy_decoder_input], numpy_decoder_output)

def _encode_segmented_corpus(
    encoded_inputs, encoded_outputs, segmentation,
    num_input_classes, num_output_classes):
    """
    propertly prepare encoded batch entry for seq2seq processing, with segmentation information.
    Useful in generators.
    The entries are simple lists of lists of codes. The function will 
    add to them the specific codes for sentence limits (1 and 2) where needed.
    - encoded_inputs: lists of lists of entry codes (integers)
    - encoded_outputs: lists of lists of entry codes (integers)
    - segmentation: list of list of boolean indicators (0 or 1)
    - num_input_classes: number of classes of input entries
    - num_input_classes: number of classes of outputs
    - returns a complex 2-uple of values ((a,b),(c,d))
        a,b are numpy arrays for encoder and decoder inputs
        c is a numpy array for the expected output
        d is a numpy array for expected segmentation.
    """
    assert len(encoded_inputs[0]) == len(segmentation[0]), "erreur sur "+ str(encoded_inputs[0]) + " et " + str(segmentation[0])
    numpy_encoder_input = _corpus_list_to_np(encoded_inputs, num_classes=num_input_classes)
    # base for the decoder...
    numpy_decoder_in_out = _corpus_list_to_np(encoded_outputs, num_classes=num_output_classes)
    numpy_decoder_input = numpy_decoder_in_out[:,:-1,:]
    numpy_decoder_output = numpy_decoder_in_out[:,1:,:]
    numpy_segmentation = np.array(sequence.pad_sequences(segmentation, padding='post'), dtype=float)    
    return ([numpy_encoder_input, numpy_decoder_input], [numpy_decoder_output, numpy_segmentation])

class BatchCorpus:
    """
    Un corpus - fourni en batch - pour des traitements de types seq2seq

    POUR L'INSTANT, en entrée on a un vocabulaire, et en sortie on a des caractères,
    mais ça va changer.

    Si c'est intéressant, on peut envisager de fournir un filtre.  
    """
    def __init__(self, vocabulary, src, tgt, batch_size):
        # lit la taille du corpus.
        with open(src) as f0:
            self._corpus_size = sum(1 for _ in f0)
        self._batch_size = batch_size
        self._vocabulary = vocabulary
        self._src = src
        self._tgt = tgt

    def corpus_size(self):
        return self._corpus_size

    def batch_size(self):
        return self._batch_size
    
    def vocabulary(self):
        return self._vocabulary

    def steps_per_epoch(self):
        if self.corpus_size() % self.batch_size() == 0 :
            steps_per_epoch = self.corpus_size() / self.batch_size()
        else:
            steps_per_epoch= (self.corpus_size() / self.batch_size()) + 1
        return steps_per_epoch

    
    def _load_target_corpus(self):
        """
        Reads the target corpus, understood as a list of characters.

        """
        return _load_character_corpus(self._tgt, space_separated= True)

    def no_batch(self):
        """
        Returns a no-batch view of the corpus. Can be used for validation corpora.
        """
        list_in = self.vocabulary().encode_file(self._src)
        list_out = self._load_target_corpus()
        return _encode_corpus_for_seq2seq(list_in, list_out,
                                    num_input_classes=self.vocabulary().size(), num_output_classes=256)

    def generator(self, epochs):
        """
        returns a generator function for this corpus.
        """
        def _mygenerator():
            def extract(a_list, indexes):
                    return operator.itemgetter(*indexes)(a_list)
            # Reads and encode the whole corpus in memory. We can do it.
            list_in = self.vocabulary().encode_file(self._src) # List of lists of codes...
            list_out = self._load_target_corpus() # list of character codes
            epoch_count= 0
            while True:
                    indexes= list(range(0,self.corpus_size()))
                    random.shuffle(indexes)
                    for i in range(0, self.corpus_size(), self.batch_size()):
                            batch_indexes = indexes[i:i+self.batch_size()]                        
                            batch_in = extract(list_in, batch_indexes)
                            batch_out= extract(list_out, batch_indexes)
                            # Note: limits have been added when reading the corpus.
                            # TODO : pass or store somewhere the size of the output character set.
                            res= _encode_corpus_for_seq2seq(
                                    encoded_inputs= batch_in, encoded_outputs= batch_out, 
                                    num_input_classes=self.vocabulary().size(), num_output_classes=256)
                            yield res
                    if epochs != None and epoch_count > epochs:
                            break
                    else:
                            epoch_count = epoch_count + 1
        return _mygenerator()    
    

def prepare_or_load_model(fname, model_generator):  
    '''
    charge un modèle depuis le disque dur, ou crée un nouveau modèle si 
    aucun n'est disponible.
    - model_file_name : le fichier h5 du modèle
    - model_generator : la fonction qui crée un nouveau modèle si nécessaire.
    - retourne le chemin du modèle, et le modèle lui-même.
    '''
    if os.path.isfile(fname) :
        modele = load_model(fname)
    else:
        modele= model_generator()
    return fname, modele



def create_and_train(
    fname,
    model_generator,
    batch_corpus,
    validation_data = None,
    period=5, 
    num_epochs=20,
    ):
    '''
        Crée et entraîne un réseau en utilisant un générateur pour lire le corpus.
        Permet de gérer de plus gros corpus.
        - fname : fichier qui contient/contiendra le modèle
        - model_generator : une fonction qui retourne un nouveau modèle, si le fichier fname n'existe pas.
        - batch_corpus : un objet de classe BatchCorpus
        - validation_data : le corpus de validation ; voir la doc de Model.fit_generator() pour specifications.
        - period : la période de sauvegarde
        - retourne le modèle créé.

        les données de validation 
    '''   
    
    fname, model= prepare_or_load_model(fname, model_generator)
    
    callbacks=[
        ModelCheckpoint(filepath=fname, verbose=1, 
            monitor='val_loss', save_best_only=True,period=period)
    ]

    corpus_generator = batch_corpus.generator(num_epochs)

    model.fit_generator(
        generator= corpus_generator,
        steps_per_epoch= batch_corpus.steps_per_epoch(),
        validation_data= validation_data,
        epochs=num_epochs,
        callbacks= callbacks
    )
    return model
    
class CorpusSegmentation:
    """
    Un corpus pour la tâche couplée segmentation/translittération.
    """
    def __init__(self, vocabulary, src, tgt, batch_size):
        # lit la taille du corpus.
        with open(src) as f0:
            self._corpus_size = sum(1 for _ in f0)
        self._batch_size = batch_size
        self._vocabulary = vocabulary
        self._src = src
        self._tgt = tgt

    def corpus_size(self):
        return self._corpus_size

    def batch_size(self):
        return self._batch_size
    
    def vocabulary(self):
        return self._vocabulary

    def steps_per_epoch(self):
        if self.corpus_size() % self.batch_size() == 0 :
            steps_per_epoch = self.corpus_size() / self.batch_size()
        else:
            steps_per_epoch= (self.corpus_size() / self.batch_size()) + 1
        return steps_per_epoch

    
    def _load_target_corpus(self):
        """
        Reads the target corpus, understood as a list of characters.
        """
        return _load_character_corpus(self._tgt, space_separated= True)

    def no_batch(self):
        """
        Returns a no-batch view of the corpus. Can be used for validation corpora.
        """
        list_in = self._build_list_in()
        segment_out = self._load_segment_out()
        list_out = self._load_target_corpus()
        return _encode_segmented_corpus(list_in, list_out, segmentation=segment_out,
                                    num_input_classes=self.vocabulary().size(), num_output_classes=256)

    def _build_list_in(self):
        return self.vocabulary().encode_file(self._src, filter=lambda s: s.replace(" _",""))
        
    def _load_segment_out(self):
        """
        build lists of 0 and 1, depending on whether the signs are end-of-words or not.
        BOS and EOS are both understood as word limits.
        """
        res= [] 
        regexp = re.compile(r" _[_ ]*")
        with open(self._src, encoding="utf-8") as f:
            for l in f:
                l = l.strip(" \n\t") # On nettoie l.
                l1 = (l.replace(" _","")).split() # l1 = l sans les séparations.
                l = regexp.sub(' _ ', l).strip(" ")
                encoded_line = [[1]] # because BOS 
                previous_token = '_'
                for token in l.split(' '):
                    if token == '_' :
                        encoded_line += [[1]]
                    elif previous_token != '_' :
                        encoded_line += [[0]]
                    previous_token = token
                if previous_token != '_':
                    encoded_line += [[0]]
                encoded_line += [[1]] # EOS.
                assert len(encoded_line) == len(l1) + 2, "erreur sur " + l
                res.append(encoded_line)
        return res

    def generator(self, epochs):
        """
        returns a generator function for this corpus.
        """
        def _mygenerator():
            def extract(a_list, indexes):
                    return operator.itemgetter(*indexes)(a_list)
            # Reads and encode the whole corpus in memory. We can do it.
            list_in = self._build_list_in() # List of lists of codes...
            list_out = self._load_target_corpus() # list of character codes
            seg_out = self._load_segment_out()
            epoch_count= 0
            while True:
                    indexes= list(range(0,self.corpus_size()))
                    random.shuffle(indexes)
                    for i in range(0, self.corpus_size(), self.batch_size()):
                            batch_indexes = indexes[i:i+self.batch_size()]                        
                            batch_in = extract(list_in, batch_indexes)
                            batch_out= extract(list_out, batch_indexes)
                            batch_out_seg = extract(seg_out, batch_indexes)
                            # Note: limits have been added when reading the corpus.
                            # TODO : pass or store somewhere the size of the output character set.
                            res= _encode_segmented_corpus(
                                    encoded_inputs= batch_in, encoded_outputs= batch_out, 
                                    segmentation=batch_out_seg,
                                    num_input_classes=self.vocabulary().size(), num_output_classes=256)
                            yield res
                    if epochs != None and epoch_count > epochs:
                            break
                    else:
                            epoch_count = epoch_count + 1
        return _mygenerator()    
