# Encoder/Decoder architecture for automated transliteration.

# We use a global attention system, but the architecture is relatively modular
# (the reason we wrote the code ourselves in the first place)
#
# Naming conventions :
# - encoder LSTM layers are called hs0, hs1
# - decoder LSTM layers are called ht0, ht1
#
# Regarding states :
# 
# - output states  : LSTM name + _out
# - inner states : LSTM name + _h ou _c


import random
import numpy
import os
import sys
import re
import matplotlib.pyplot as plt

from keras.callbacks import ModelCheckpoint
from keras.layers import LSTM, Masking, Dense, Dot, Softmax, Concatenate, Bidirectional
from keras.models import Input, load_model, Model
from keras.preprocessing import sequence
from keras.utils import to_categorical, plot_model

from translit_lib.corpus_utils import computeRepresentation, decodeRepresentation, prepare_or_load_model
from translit_lib.corpus_utils import BatchCorpus, create_and_train
from translit_lib.corpus_utils import clean_mdc_input, encode_mdc_string
from translit_lib.vocabulary import Vocabulary
from translit_lib.beam import Beam

class Encoder:
    """
    todo : use a more "kera-like" architecture ?

    Notable fields for the encoder :
    input/output :
    - input : input layer (BATCH x LENGTH_I x input_size)
    - hs0_out : output of the first level of LSTM (BATCH x LENGTH_I x hidden)
    - hs0_h
    - hs0_c : hidden layers for the first level LSTM
    - hs1_out : output of the second level of lstm (BATCH x LENGTH_I x hidden)
    - hs1_h
    - hs1_c : hidden layers for the second level LSTM (and so, of the encoder)
        
    For attention, etc... use hs1_out

    - hs0 : first lstm
    - hs1 : second lstm

    """

    def __init__(self, hidden=300, input_size=256):
        """
        build an encoder.
        """
        self.hidden = hidden
        self.input_size = input_size
        self.input= Input((None,input_size), name='encoderInput') # BATCH x LENGTH_I x input_size
        # La couche de masque est très très importante. Essayer en la supprimant.
        # On retourne le dernier état en plus du résultat, pour le réinjecter comme premier 
        # état dans le décodeur.
        self.mask= Masking(name='masking_1')
        self.maskOutput= self.mask(self.input)    # BATCH x LENGTH_I x input_size
        self.hs0= LSTM(hidden, return_sequences=True, return_state=True, 
            name='hs0', implementation=2, dropout=0.3, recurrent_dropout=0.3)
        self.hs0_out, self.hs0_h, self.hs0_c= self.hs0(self.maskOutput) # BATCH x LENGTH_I x hidden

        # ajouter une seconde couche d'encodeurs ?
        self.hs1= LSTM(hidden, return_sequences=True, return_state=True, 
            name='hs1', implementation=2, dropout=0.3, recurrent_dropout=0.3)
        self.hs1_out, self.hs1_h, self.hs1_c= self.hs1(self.hs0_out)  # BATCH x LENGTH_I x hidden
        self.output = self.hs1_out

    def load_model_from_file(self, fname):
        """
        Crée un réseau "stand-alone" et charge ses poids à partir d'un fichier.
        Le réseau en question sera utilisé en collaboration avec un décodeur.

        La sortie aura la structure suivante:
        self.hs0_out, self.hs0_h, self.hs0_c, self.hs1_out, self.hs1_h, self.hs1_c

        Notez que la sortie "intéressante" est self.hs1_out.
        """ 
        
        model = Model(
            inputs= [self.input],
            outputs=[self.hs0_out, self.hs0_h, self.hs0_c, self.hs1_out, self.hs1_h, self.hs1_c]
        )
        model.load_weights(fname, by_name=True)
        return model

class BidiEncoder:
    """
    A bi-directional encoder.

    Champs notables de l'encodeur :
    Entrées/sorties :
    - input : couche d'entrée (BATCH x LENGTH_I x input_size)
    - hs0_out : sortie du premier niveau de lstm (BATCH x LENGTH_I x hidden)
    - hs0_h
    - hs0_c : couches cachées finales du premier LSTM
    - hs1_out : sortie du premier niveau de lstm (BATCH x LENGTH_I x hidden)
    - hs1_h
    - hs1_c : couches cachées finales du second LSTM (et donc du réseau en entier)
    
    La couche qui sert habituellement pour l'attention, etc... est la couche hs1_out

    Couches du réseau:
    - hs0 : premier lstm
    - hs1 : second lstm

    """

    def __init__(self, hidden=300, input_size=256):
        """
        Construit un encodeur bi-directionnel.
        - hidden : la dimension des couches utilisées ; attention, pour rester comparable avec 
                   les réseaux non-bi-directionnels, ces dimensions sont ensuite divisées par deux...        
        """
        self.hidden = hidden
        self.input_size = input_size
        self.input= Input((None,input_size), name='encoderInput') # BATCH x LENGTH_I x input_size
        # La couche de masque est très très importante. Essayer en la supprimant.
        # On retourne le dernier état en plus du résultat, pour le réinjecter comme premier 
        # état dans le décodeur.
        self.mask= Masking(name='masking_1')
        self.maskOutput= self.mask(self.input)    # BATCH x LENGTH_I x input_size
        
        # Note sur les réseaux bidirectionnels
        self.hs0= Bidirectional(layer=
            LSTM(hidden//2, return_sequences=True, return_state=True, 
                name='hs0_fw', implementation=2, dropout=0.3, recurrent_dropout=0.3),
            name= 'hs0')
        # En lisant le code de Bidirectional, il retourne d'abord les états forward, puis 
        # les états backwards.
        
        self.hs0_out, h0f, c0f, h0b, c0b = self.hs0(self.maskOutput) # BATCH x LENGTH_I x hidden 
        self.concat = Concatenate(name="concat")
        self.hs0_h = self.concat([h0f, h0b])
        self.hs0_c = self.concat([c0f, c0b])
        
        # self.hs0_out, self.hs0_h, self.hs0_c= self.hs0(self.maskOutput) 

        # Note pour la suite: les états récupérés par le décodeur ? concaténation du dernier état forward
        # et du premier état backward ? 
        
        # ajouter une seconde couche d'encodeurs ?
        self.hs1= Bidirectional(layer=
            LSTM(hidden//2, return_sequences=True, return_state=True, 
                name='hs1_fw', implementation=2, dropout=0.3, recurrent_dropout=0.3),
            name= 'hs1')
        self.hs1_out, h1f, c1f, h1b, c1b= self.hs1(self.hs0_out)  # BATCH x LENGTH_I x hidden 
        self.hs1_h = self.concat([h1f, h1b])
        self.hs1_c = self.concat([c1f, c1b])
        
        self.output = self.hs1_out

    def load_model_from_file(self, fname):
        """
        Crée un réseau "stand-alone" et charge ses poids à partir d'un fichier.
        Le réseau en question sera utilisé en collaboration avec un décodeur.

        La sortie aura la structure suivante:
        self.hs0_out, self.hs0_h, self.hs0_c, self.hs1_out, self.hs1_h, self.hs1_c

        Notez que la sortie "intéressante" est self.hs1_out.
        """ 

        model = self.build_model()
        model.load_weights(fname, by_name=True)
        return model
        
    def build_model(self):
        model = Model(
            inputs= [self.input],
            outputs=[self.hs0_out, self.hs0_h, self.hs0_c, self.hs1_out, self.hs1_h, self.hs1_c]
        )
        return model

class Decoder():
    """
    Construit un décodeur avec attention.
    """
    def __init__(self, hidden=300, encoder_input_size=256, 
                decoder_input_size=256, encoder_output=None, 
                encoder_states=None):
        """
        Initialise un décodeur.
        En exploitation, l'encodeur et le décodeur seront dissociés. Dans ce cas, laisser encoder_output et encoder_states à None ; 
        le décodeur créera ce dont il a besoin.

        - encoder_output : la sortie de l'encodeur - uniquement si on veut combiner encodeur et décodeur en un seul réseau.
        - encoder_states : liste des états finaux des LSTM de l'encodeur. Pour l'instant limitées à deux LSTM

        """

        if (encoder_output == None):
            self._prepare_entries(hidden)
            encoder_output = self.encoder_output
            hs0_h, hs0_c, hs1_h, hs1_c = self.hs0_h, self.hs0_c, self.hs1_h, self.hs1_c
            # états dans le cas du décodeur en exploitation...
            HT_return_state= True
        else:
            hs0_h, hs0_c, hs1_h, hs1_c = encoder_states
            HT_return_state= False

        self.input = Input((None,decoder_input_size), name='decoderInput')  # BATCH x LENGTH_O x output_size
        self.decoder_input_mask = Masking(name='masking_2')
        decoder_input_mask_out = self.decoder_input_mask(self.input) # BATCH x LENGTH_O x output_size

        # NOTER L'UTILISATION DES DEUX ETATS ISSUS DE L'ENCODEUR.
        self.ht0 = LSTM(hidden, name='ht0', 
            return_sequences= True,
            return_state= HT_return_state,
            implementation=2, dropout=0.3, recurrent_dropout=0.3)

        res0 = self.ht0(decoder_input_mask_out, initial_state=[hs0_h,hs0_c])
        if HT_return_state:
            self.ht0_out, self.ht0_h, self.ht0_c = res0
        else:
            self.ht0_out = res0
        
        self.ht1 = LSTM(hidden, name='ht1', 
            return_sequences= True, 
            return_state= HT_return_state,
            implementation=2, dropout=0.3, recurrent_dropout=0.3)

        res1 = self.ht1(self.ht0_out, initial_state=[hs1_h,hs1_c]) # BATCH x LENGTH_O x hidden
        if HT_return_state:
            self.ht1_out, self.ht1_h, self.ht1_c = res1
        else:
            self.ht1_out = res1
        
        # Couches d'attention

        # calcul de score
        # score(h_t, h_s) = (h_t^T W_a h_s) 
        self.score_dense = Dense(hidden, name='W_a') 
        score_dense_out = self.score_dense(encoder_output) # BATCH x LENGTH_I x hidden
        
        self.scoreDot = Dot(axes=(2,2), normalize=False, name='scoreDot')
        score_out = self.scoreDot([score_dense_out,self.ht1_out]) # BATCH x LENGTH_I x LENGTH_O

        # calcul de a_t
        self.at = Softmax(axis= 1, name="attention") # on normalise le long de l'axe LENGTH_I
        self.at_out = self.at(score_out) # BATCH x LENGTH_I x LENGTH_O

        # Calcul de c_t
        self.ct = Dot(axes=(1,1), name='c_t')
        # dimensions des données:
        # at_out : BATCH x LENGTH_I x LENGTH_O
        # hs1_out : BATCH x LENGTH_I x hidden
        # hésitation sur les dimensions du résultat... faut-il utiliser TimeDistributed ?
        # Au vu de model.png, mon estimation est bonne...
        ct_out = self.ct([self.at_out, encoder_output]) # BATCH x LENGTH_0 x hidden ? 

        # h~t= tanh(Wc[ct; ht])
        self.concat = Concatenate(axis=-1, name= "concatenate_1")
        concat_out = self.concat([ct_out, self.ht1_out]) # BATCH x LENGTH_O x (hidden + hidden)
        self.w_c = Dense(hidden,  name="W_c", activation= "tanh")
        w_c_out =  self.w_c(concat_out) # BATCH x LENGTH_0 x hidden
        # Le nom de la couche "softmax" est très très très mal choisi...
        self.soft1 = Dense(decoder_input_size, name="softmax", activation="softmax")
        self.output = self.soft1(w_c_out)

    def _prepare_entries(self, hidden):
        """
        Dans le cas où le décodeur est 'stand alone', prépare les couches d'entrée.
        """                    
        # En phase d'apprentissage, le décodeur est directement relié à 
        # l'encodeur.
        
        # En phase d'exploitation, on va séparer les deux.
        # Voici la configuration en exploitation :
        # Sortie du décodeur (couche hs1_out) : sert au calcul de l'attention.
        self.encoder_output = Input((None, hidden), name='encoderOutput')
        # états initiaux (pas dans le décodeur d'apprentissage, où ils sont issus de l'encodeur)
        self.hs0_h = Input((hidden,), name='hs0_h')
        self.hs0_c = Input((hidden,), name='hs0_c')
        self.hs1_h = Input((hidden,), name='hs1_h')
        self.hs1_c = Input((hidden,), name='hs1_c')

    def load_model_from_file(self, fname, attention_output=False):
        """
        Crée un réseau "stand-alone" et charge ses poids à partir d'un fichier.
        """ 
        inputs = [self.input, self.encoder_output, 
            self.hs0_h, self.hs0_c, self.hs1_h, self.hs1_c]
        outputs = [self.output, 
                    self.ht0_out, self.ht0_h, self.ht0_c, 
                    self.ht1_out, self.ht1_h, self.ht1_c]
        if attention_output:
            outputs = outputs + [self.at_out]
        model= Model(inputs=inputs, outputs=outputs)
        model.load_weights(fname, by_name=True)
        return model

class EncoderDecoder:
    """
    Encodeur/Décodeur avec attention générale, cf. Luong et al.
    """

    def __init__(self,hidden=300, encoder_input_size=256, decoder_input_size=256, bidi=False):
        if bidi:
            self.encoder = BidiEncoder(hidden= hidden, input_size= encoder_input_size)
        else:
            self.encoder = Encoder(hidden=hidden, input_size= encoder_input_size)

        self.decoder = Decoder(hidden= hidden, 
            encoder_input_size=encoder_input_size, decoder_input_size= decoder_input_size, 
            encoder_output=self.encoder.output, 
            encoder_states=[
                self.encoder.hs0_h, 
                self.encoder.hs0_c, 
                self.encoder.hs1_h, 
                self.encoder.hs1_c, 
            ])
        
    def encoder_input(self):
        return self.encoder.input

    def decoder_input(self):
        return self.decoder.input

    def decoder_output(self):
        return self.decoder.output

    def build_model(self):
        model = Model(inputs=[self.encoder_input(), self.decoder_input()], outputs=[self.decoder_output()])
        model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['acc'])
        return model

class EncoderDecoderSegmentation:
    """
    Encodeur-décodeur plus segmentation.
    """
   
    def __init__(self,hidden=300, encoder_input_size=256, 
                decoder_input_size=256,
                segment_dim=300):
        self.encoder = Encoder(hidden=hidden, input_size= encoder_input_size)
        self.decoder = Decoder(hidden= hidden, 
            encoder_input_size=encoder_input_size, decoder_input_size= decoder_input_size, 
            encoder_output=self.encoder.output, 
            encoder_states=[
                self.encoder.hs0_h, 
                self.encoder.hs0_c, 
                self.encoder.hs1_h, 
                self.encoder.hs1_c, 
            ])

        self.segment_dense = Dense(units=segment_dim, name="segment_1", activation='relu')
        self.segment_dense_inter = self.segment_dense(self.encoder.hs1_out)
        self.segment_dense = Dense(units=1, name="segment_2", activation='relu')        
        self.segment_dense_out = self.segment_dense(self.segment_dense_inter)

        
    def encoder_input(self):
        return self.encoder.input

    def decoder_input(self):
        return self.decoder.input

    def decoder_output(self):
        return self.decoder.output

    def segment_output(self):
        return self.segment_dense_out

    def build_model(self):
        model = Model(
            inputs=[self.encoder_input(), self.decoder_input()],
            outputs=[self.decoder_output(), self.segment_output()] 
        )
        #losses = [, ]        
        losses = {
            'softmax' : 'categorical_crossentropy', 
            'segment_2': 'mean_squared_error'
        }
        # essayer aussi 2.0 et 1.0 ?
        loss_weights= {
            'softmax' : 1.0, 
            'segment_2': 1.0
        }
        model.compile(optimizer='adam', 
            loss= losses, loss_weights= loss_weights,
            metrics=['acc'])
        return model

class EncoderDecoderBidiSegmentation:
    """
    Encodeur-décodeur bidirectionnel plus segmentation.
    """
   
    def __init__(self,hidden=300, encoder_input_size=256, 
                decoder_input_size=256,
                segment_dim=300):
        self.encoder = BidiEncoder(hidden=hidden, input_size= encoder_input_size)
        self.decoder = Decoder(hidden= hidden, 
            encoder_input_size=encoder_input_size, decoder_input_size= decoder_input_size, 
            encoder_output=self.encoder.output, 
            encoder_states=[
                self.encoder.hs0_h, 
                self.encoder.hs0_c, 
                self.encoder.hs1_h, 
                self.encoder.hs1_c, 
            ])

        self.segment_dense = Dense(units=segment_dim, name="segment_1", activation='relu')
        self.segment_dense_inter = self.segment_dense(self.encoder.hs1_out)
        self.segment_dense = Dense(units=1, name="segment_2", activation='relu')        
        self.segment_dense_out = self.segment_dense(self.segment_dense_inter)

        
    def encoder_input(self):
        return self.encoder.input

    def decoder_input(self):
        return self.decoder.input

    def decoder_output(self):
        return self.decoder.output

    def segment_output(self):
        return self.segment_dense_out

    def build_model(self):
        model = Model(
            inputs=[self.encoder_input(), self.decoder_input()],
            outputs=[self.decoder_output(), self.segment_output()] 
        )
        losses = {
            'softmax' : 'categorical_crossentropy', 
            'segment_2': 'mean_squared_error'
        }
        # essayer aussi 2.0 et 1.0 ?
        loss_weights= {
            'softmax' : 1.0, 
            'segment_2': 1.0
        }
        model.compile(optimizer='adam', 
            loss= losses, loss_weights= loss_weights,
            metrics=['acc'])
        return model


def save_separate_models(encoder_decoder_model_path,
     vocabulary_path= "vocab-in.txt",
     encoder_model_path="encoder_model.h5", 
     decoder_model_path="decoder.model.h5") :
    """
     Takes an encoder-decoder model, and saves two separated model, one for encoder and one for decoder.
     It should ease the export to other systems, like java.
     - encoder_decoder_model_path: the original h5 model path
     - vocabulary_path : the entry layer vocabulary (not saved, but its size is important)
     - encoder_model_path: the new encoder model path
     - decoder_model_path: the new decoder model path
    """
    vocabulary = Vocabulary()
    vocabulary.load(vocabulary_path)

    encoder = Encoder(input_size=vocabulary.size())
    encoder_model = encoder.load_model_from_file(encoder_decoder_model_path)
    encoder_model.save(encoder_model_path)

    decoder = Decoder()
    decoder_model = decoder.load_model_from_file(encoder_decoder_model_path)
    decoder_model.save(decoder_model_path)

class _BeamNode:
    """
    A node in a beam search. Utility class we will probably hide later.
    """
    def __init__(self, text, last_char, h0, c0, h1, c1, attention=None):
        self.text = text
        self.last_char = last_char
        self.h0 = h0
        self.h1 = h1
        self.c0 = c0
        self.c1 = c1
        self.attention = attention

    def closed(self):
        return self.last_char == 2

    def states(self):
        return (self.h0,self.c0,self.h1,self.c1)

def _agregate_values(vals):
    """
    takes a list of couple (value, pos) and concatenates consecutive values.
    The result will also be a list of couples (val, pos), but the values will be discontinuous.
    """
    if len(vals) == 0:
        return []
    else:
        res = []        
        concatenated_value, prev_pos = vals[0]
        start_pos = prev_pos
        for (value, pos) in vals[1:] :
            if pos == prev_pos + 1:
                concatenated_value += value            
            else:
                res += [(concatenated_value, start_pos)]
                concatenated_value = value
                start_pos = pos
            prev_pos = pos
        res += [(concatenated_value, start_pos)]
        return res

class AttentionTable:
    """
    A user-friendly representation of attention.
    This is not used in the transliteration process itself,
    but it's a way to exploit the result of the said process.
    """

    def __init__(self, attention, glyphs, letters):
        self._attention = numpy.copy(attention)
        self._glyphs = glyphs
        self._letters = letters

    def get_number_of_glyphs(self):
        return len(self._glyphs)
    
    def get_glyph(self, pos):
        return self._glyphs[pos]

    def get_number_of_letters(self):
        return len(self._letters)
    
    def get_letter(self, pos):
        return self._letters[pos]

    def get_attention_for(self, glyph_pos, letter_pos):
        return self._attention[glyph_pos, letter_pos]
    
    def get_argmax_for_letter(self, letter_pos):
        """
        gets the position of the sign which best explains a specific letter.
        """
        return numpy.argmax(self._attention,axis=0)[letter_pos]

    
    def sign_values(self, threshold = 0.9):
        """
        Extract sign information from a sentence analysis.
        More specifically, will return a list of the following form :
        each element will be a n-uple with :
        A) the (hieroglyphic) sign position
        B) the sign in position i
        C) An ordered list of all transliteration signs for which sign i receives the most attention (first-class value)
            Each element of the list is a transliteration sign, plus a position in the original
            we might decide to aggregate the consecutive elements
        D) A similar list, but from a list of signs for which i is needed to reach a 99% coverage of the transliteration sign's 
            attention (secondary values)
        Note that C and D might be empty.
        """
        res = []
        signVals={}
        secondaryVals={}
        for j in range(self.get_number_of_letters()):
            # Most significant sign
            i_max = self.get_argmax_for_letter(j)
            if i_max in signVals:
                r = signVals[i_max]      
            else:
                r = []
            r += [(self.get_letter(j), j)]
            signVals[i_max] = r        

            # Compute secondary values :
            ordered = numpy.argsort(self._attention[:,j])[::-1]
            total = self._attention[ordered[0],j]
            for k in range(1,len(ordered)):
                if total >= threshold:
                    break
                else:
                    i = ordered[k]
                    att_i = self._attention[i,j]
                    total += att_i
                    if i in secondaryVals:
                        r = secondaryVals[i]
                    else:
                        r = []
                    r += [(self.get_letter(j), j)]
                    secondaryVals[i] = r
        for k in range(self.get_number_of_glyphs()):
            glyph = self.get_glyph(k)
            if k in signVals:
                part1 = _agregate_values(signVals[k])
            else:
                part1 = None
            if k in secondaryVals:
                part2 = _agregate_values(secondaryVals[k])
            else:
                part2 = None
            res += [(k, glyph, part1, part2)]
        return res

    def _sign_to_letter_values(self, glyph_pos, threshold):
        """
        Returns the letters whose attention to a given glyph is greater than threshold.
        Returns a list of couples (letter, pos).
        """
        res = []
        for i in range(self.get_number_of_letters()):
            a =  self.get_attention_for(glyph_pos, i)
            if (a >= threshold):
                res += [ (self.get_letter(i), i) ]
        return res

    def sign_to_letters_values(self, threshold=0.1):
        """
        For each glyph, returns the letters whose attention to it is greater than threshold.
        Results is a list of n-uplets
            (glyph, glyph_position, concatenated values), 
            where concatenated values is a list of couples (string, position)
        """
        res = []
        for i in range(self.get_number_of_glyphs() - 1): # O-based indexes!
            glyph = self.get_glyph(i)
            vals = self._sign_to_letter_values(i, threshold)
            if (len(vals) > 0) :
                concatenated_values = _agregate_values(vals)
                res += [(glyph, i, concatenated_values)]
        return res


    def get_glyph_indexes(self, glyph):
        return [i for i in range(self.get_number_of_glyphs()) if self.get_glyph(i) == glyph]

class Transliterator:
    """
    An object which knows how to transliterate text.

    The transliterate method will return "the" best solution.
    The rest of the data produced will be accessible from getters.
    """

    def __init__(self, vocabulary, encoder_model, decoder_model,attention_output=False):
        """
        Create a transliterator.
        Note that both the encoder and the decoder are keras models. If you want to use
        Classes such as Encoder or Decoder, get the models with load_model_from_file.
        - vocabulary : the vocabulary used to encode the input
        - encoder_model : a model object for the encoder ; typically loaded from a h5 file
        - decoder_model : a model object for the encoder ; typically loaded from a h5 file
        - attention_output : does the decode output includes a last value which is the attention ?
        Nota: this is a redundancy which might be avoided if we passed encoder and decoder objets and not simple keras 
            models.
        """
        super().__init__()
        self.vocabulary = vocabulary
        self.encoder_model = encoder_model
        self.decoder_model = decoder_model
        self.attention_output = attention_output
    
    def get_beams(self):
        """
        return the best beams produced by the latest search.
        """
        return self.beams

    def get_attention_table(self):
        """
        Returns a usable table for attention.

        """
        glyphs = ["<BOS>"] + self._input_string.split(" ") + [ "<EOS>"]
        letters = [chr(c) for c in self._best_output[1:]]
        return  AttentionTable(numpy.transpose(self._attention), glyphs, letters)
        
    def plot_attention(self, picture = None):
        """
        Plot the attention.
        If a file is specified, the attention will be saved there.
        Else, it will simply be displayed.
        """
        input_list = self._input_string.split(" ")
        plt.matshow(numpy.transpose(self._attention), cmap='binary')

        plt.yticks(range(len([1]+ input_list)+1), [1]+ input_list+[2])
        x_labels = [chr(c) for c in self._best_output[1:]]  
        plt.xticks(range(len(self._best_output[1:])), x_labels)
        if picture:
            plt.savefig(picture)
        else:
            plt.show()

   
    def transliterate(self, input_string, beam_width=5):
        """
        Beam search transliteration.
        """
        # keep last input string for attention display.
        # TODO : make a better system for pre-processing, in order to avoid code duplication.
        # (kind of : a preprocessor class, which will give access to all relevant data.)
        self._input_string = clean_mdc_input(input_string)
        encoded_input = encode_mdc_string(self._input_string, vocabulary=self.vocabulary)
        max_size = 4 * len(self._input_string.split(" ")) # Rough estimation from training corpus of the max input to output ratio.
        # Encode the input
        _, h0, c0, hs1_out, h1, c1 = self.encoder_model.predict(encoded_input)
        # Initialize current set of beams.
        parent_beam = Beam(width=beam_width) # the graph
        attention = [] # we don't have attention in the beginning
        start_node = _BeamNode([1], 1, h0, c0, h1, c1, attention)
        parent_beam.add(0, start_node)

        current_output_size = 0
        finished = False # replace later...

        while not finished and current_output_size < max_size:
            current_output_size += 1
            has_opened_node = False # does the beam contain a node which can be expanded ?
            current_beam = Beam(width=beam_width) # new beam set will be built.
            for (cost,node) in parent_beam.values() : # We try to expand the old beams:
                if  node.closed() :
                       current_beam.add(cost, node) # already closed nodes are simply copied over.
                else:
                    # Extend node :
                    previous_attention = node.attention
                    has_opened_node = True # There was an opened node, search is not finished
                    # Compute next state :
                    encoded_output = computeRepresentation([[node.last_char]])
                    decoder_input = [encoded_output, hs1_out, *(node.states())]
                    prediction = self.decoder_model.predict(decoder_input)
                    if self.attention_output:
                        htilde_out, _, h0, c0, _, h1, c1, last_attention = prediction
                        attention = previous_attention + [numpy.ravel(last_attention)]
                    else:
                        htilde_out, _, h0, c0, _, h1, c1 = prediction
                    # Select the most likely letters :
                    distribution = htilde_out[0][-1]
                    candidate_indexes = numpy.argpartition(-distribution, beam_width)[:beam_width]
                    # Add the corresponding nodes to the current beam.
                    for c in candidate_indexes:
                        new_node = _BeamNode(node.text + [c], c, h0, c0, h1, c1, attention)
                        current_beam.add((cost*len(node.text) + distribution[c]) / (len(node.text) + 1), new_node)
            parent_beam = current_beam
            if not has_opened_node:
                finished= True
        # Keep best beams if we want to see them.
        self.beams = []
        for (cost, node) in parent_beam.values():
                translit = "".join(([chr(c) for c in node.text]))[1:-1].replace("_", " ")
                self.beams.append(translit)
        
        # Store last attention results.
        if self.attention_output:
            self._attention = parent_beam.best_node().attention
        else :
            self._attention = None

        output_list = parent_beam.best_node().text
        res = "".join(([chr(c) for c in output_list]))[1:-1].replace("_", " ")
        self._best_output = parent_beam.best_node().text
        return res


class Segmentor:
    """
    A trained segmentor, made from a encoder-decoder for bi-directionnal segmentation.
    """
    pass

    def __init__(self, h5_file, vocabulary):
        """
        Initialize a segmentor.
        - h5_file : a file from a segmentation-based model 
        - vocabulary : the vocabulary linked with this model        
        """
        tmp_model = load_model(h5_file)
        input_layer = tmp_model.get_layer("encoderInput")
        output_layer = tmp_model.get_layer("segment_2")
        self._model = Model(inputs=[input_layer.input], outputs=[output_layer.output])
        self._vocabulary = vocabulary
        # Nothing done for attention output yet
    
    def segment(self, string):
        """
        Returns the segmentation result for a given input.
        - string : a string containing space-separated MDC codes 
        - returns : a list of couples (MDC code, separator likelyhood)
        """
        string = clean_mdc_input(string)
        codes = string.split(" ")
        input = encode_mdc_string(string, self._vocabulary)
        output = self._model.predict(input)
        print(output)
        res = []
        for i in range(len(codes)):
            c = codes[i]
            v = output[0,i+1,0] # i+1 because 0 = sentence begin. 
            res += [(c,v)]
        return res
    