"""
A vocabulary for input (or maybe output).

This class will keep an word-to-id index and allow one to transform a symbolic input into
indexes a neural network will be able to use.

It reserves specific codes for start of text, end of text, and unknown entries.
"""
from pathlib import Path


class Vocabulary(object):
    """
    An index for words in a vocabulary.
    indexes 0 to 9 are reserved values.
    Currently:
        0: used for masking
        1: start of text
        2: end of text
        3: unknown entry.
    """

    def __init__(self):
        """
        init the vocabulary from a file.
        """
        self.MASK = 0
        self.BOS = 1
        self.EOS = 2
        self.UNKNOWN = 3

        self._w2i = {}
        self._i2w= {}
        self._record_reserved(self.MASK, "<MASK>") # 0
        self._record_reserved(self.BOS, "<BOS>") # 1
        self._record_reserved(self.EOS, "<EOS>") # 2
        self._record_reserved(self.UNKNOWN, "<UNKNOWN>") # 3
        self.last_index = 9

    def loadDefault(self):
        """
        Loads the default vocabulary included in this distribution.
        """
        base_path = Path(__file__).parent
        file_path = (base_path / "vocab-in.txt").resolve()
        self.load(str(file_path))

    def load(self, file_name):
        with open(file_name, encoding= "utf-8") as f:
            for w in f:
                w = w.strip("\n\r")
                self._add(w)

    def _record_reserved(self, index, string):
        """
        Register an entry for a reserved index (below 10)
        """
        if index >= 10 : raise "Forbidden index"
        self._w2i[string] = index
        self._i2w[index]= string


    def _add(self, string):
        self.last_index += 1
        self._w2i[string] = self.last_index
        self._i2w[self.last_index]= string

    def string_to_index(self, string, unknown_probability= 0):
        if unknown_probability > 0 and random.random() < unknown_probability:
            return self._w2i["<UNKNOWN>"]

        if string in self._w2i:
            return self._w2i[string]
        else:
            return self._w2i["<UNKNOWN>"]

    def index_to_string(self,index):
        if index in self._i2w:
            return self._i2w[index]
        else:
            return self._i2w[3]
    
    def encode_string(self, string, unknown_probability= 0, add_limits= True):
        """
        Encode a sentence, given as a space-separated sequence of words.
        For more complex separation, preprocessing will be needed.
        - unknown_percentage: randomly replace words with "UNKNOWN" with this probability
        - returns a list of codes.
        """ 
        res = self.encode_list(string.strip(" \n\r").split(" "),
             unknown_probability= unknown_probability, add_limits= add_limits)
        return res
        

    def encode_list(self, list, unknown_probability= 0, add_limits= True):
        """
        Encode a sentence, given as a list of strings.
        - unknown_percentage: randomly replace words with "UNKNOWN" with this probability
        - add_limits : add BOS and EOS marker
        """            
        if add_limits:
            list= ["<BOS>"] + list + ["<EOS>"]
        return [self.string_to_index(w, unknown_probability) for w in list]

    def encode_file(self, file_name, unknown_probability= 0, add_limits= True, filter=None):
        """
        Encode a whole file content.
        - filter: a function string -> string which will be applied to lines in order to clean them up.
        """
        res= []
        with open(file_name, encoding="utf-8") as f:
            for l in f:
                l = l.strip(" \n\t")                
                if filter != None:
                    l = filter(l)
                l = " ".join(l.split()) # removes all sequences of whitespaces
                res.append(self.encode_string(string=l, unknown_probability= unknown_probability, add_limits=add_limits))
        return res

    def decode_list(self, codes):
        """
        Transpose a list of codes into a list of strings.
        """
        return [self.index_to_string(w) for w in codes]


    def size(self):
        return self.last_index + 1


def prepareVocabulary(files, output_file):
    """
    Prepare a vocabulary file.
    A vocabulary file is simply a text file, with one word per line.
    There should be no duplicate in it.
    When reading the vocabulary, the system will use line numbers as id, starting 
    with 10 for the first line, as 0,1 and 2 are reserved codes (and we might want more codes afterwards too) 
    """
    for file_name in files:
        s= set()
        with open(file_name,encoding="utf-8") as f:
            for l in f:
                l = l.strip(" \n\t")
                for w in l.split():
                    s.add(w)
        with open(output_file, "w",encoding="utf-8") as f:
            for w in sorted(s):
                f.write(w)
                f.write('\n')
